#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser;
######################################################################

use strict;
use Cognition::DocModel::Section;
use Cognition::DocModel::XOXO;
use Cognition::DocModel::Table;
use Cognition::DocModel::Figure;
use Cognition::HTMLParser::Utils;
use Cognition::RDFModel::Subject;
use Cognition::RDFModel::Property;
use Cognition::RDFModel::Value;
use Cognition::uF::hCard;
use Cognition::uF::hAlarm;
use Cognition::uF::hEvent;
use Cognition::uF::hTodo;
use Cognition::uF::hCalendar;
use Cognition::uF::adr;
use Cognition::uF::geo;
use Cognition::uF::RelTag;
use Cognition::uF::RelLicense;
use Cognition::uF::RelEnclosure;
use Cognition::uF::xoxo;
use Cognition::uF::figure;
use Data::Dumper;
use Encode;
use HTML::Entities qw(:DEFAULT encode_entities_numeric %entity2char);
use HTML::TreeBuilder;
use HTTP::Request;
use RDF::Redland;
use URI::URL;
use XML::LibXML qw(:all);
use XML::LibXSLT;


BEGIN
{
	our $VERSION = 0.1;
	our $XDP;
	$Data::Dumper::Indent = 1;
}


sub new
{
	my $class    = shift;
	my $source   = shift;
	my $opts     = shift;
	
	my $this   = {
		'UA'       => $opts->{ua},
		'REQUEST'  => $opts->{request},
		'RESPONSE' => $opts->{response},
		'options'  => {
			'p_AUTO'       => 1,
			'p_erdf'       => 0,
			'p_rdfa'       => 1,
			'p_rdf'        => 1,
			'p_grddl'      => 1,
			'p_structure'  => 1,
			'p_uf'         => 1
		}
	};
	bless $this, $class;
	
	$this->{BASE}    = $this->{RESPONSE}->base();
	$this->{Source} = $source; 

	foreach my $o (qw(p_erdf p_grddl p_structure p_uf p_rdfa))
	{
		$this->{options}->{$o} = $opts->{$o}
			if (defined $opts->{$o});
	}

	# XML::LibXML parser.
	$Cognition::HTMLParser::XDP = XML::LibXML->new();
	
	# Attempt to parse.
	if (!eval {
		$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
	})
	
	# If parsing failed, then 
	{
		$this->{_raw_source} = $this->{Source};
		
		# Figure out encoding.
		my $enc = 'iso-8859-1';
		if ($this->{RESPONSE}->header('content-type') =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i
		|| substr($this->{Source},0,2048) =~ /\bencoding=[\"\'\s]?([a-z0-9-]+)/i 
		|| substr($this->{Source},0,2048) =~ /\bcharset=[\"\'\s]?([a-z0-9-]+)/i )
			{ $enc = lc($1); }
		
		# Decode into UTF-8.
		if ($enc ne 'utf-8')
		{
			$this->{decoded_from} = $enc;
			$this->{Source} = decode($enc, $this->{Source});
		}

		# Try parsing again.
		if (!eval {
			$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
		})
		
		# If it still fails, then
		{
			# Last ditch attempt at getting this into usable XHTML...
			$this->{Source} = $this->html2xhtml($this->{Source});
			
			# If we can't parse it yet, then die.
			if (!eval {
				$this->{DOM} = $Cognition::HTMLParser::XDP->parse_string($this->{Source});
			})
			{
				print "COULD NOT PARSE!\n\n";
				print $this->{Source}."\n";
				die;
			}
		}
	}

	# These are handy.
	Cognition::HTMLParser::Utils::PrepareXPointers($this->{DOM});

	# Parse metadata automatically if that option is enabled.
	$this->parse_metadata if ($this->{options}->{p_AUTO});

	# We're done.
	return $this;
	
} #/sub new


sub dumpText
{
	my $this = shift;
	return Dumper($this);
} #/sub dumpText


sub dumpTextClean
{
	my $this = shift;
	my $tmp  = {};
	$tmp->{DOM} = $this->{DOM};  $this->{DOM} = '...hidden...';
	# $tmp->{XP} = $this->{XP};  $this->{XP} = '...hidden...';
	
	$Data::Dumper::Sortkeys = \&_cleanKeys;
	my $txt = $this->dumpText;
	
	$this->{DOM} = $tmp->{DOM};
	# $this->{XP} = $tmp->{XP};

	return $txt;
} #/sub dumpTextClean


sub _cleanKeys
{
	my $hash = shift;
	my @keys = sort grep {/^(_subject|_id|[^_])/} keys %{ $hash };
	return \@keys;
}


sub quoted
{
	return '"'.(shift).'"';
} #/sub quoted


sub dumpRDF
{
	my $this    = shift;
	
	my %lookup;
	
	my $rv = "<?xml version=\"1.0\"?>\n"
		. "<rdf:RDF";
		
	foreach my $ns (keys %{ $this->{NS} })
	{
		my $nsurl = $this->{NS}->{$ns}->{nsurl};
		next unless (length $nsurl);
		$lookup{$nsurl} = lc($ns);
		
		$rv .= "\n\txmlns:".lc($ns)."=\"".encode_entities_numeric($nsurl)."\"";
	}
	$rv .= ">\n\t<!-- Probably more namespaces than strictly necessary. -->\n\n";
	
	foreach my $S (sort keys %{ $this->{RDF} })
	{
		my $subject = $this->{RDF}->{$S};
		$rv .= sprintf("\t<rdf:Description rdf:about=\"%s\">\n", encode_entities_numeric($subject->uri));
		foreach my $property ($subject->propertiesObjects)
		{
			next if ($property->nVals == 0);
			next unless (length $property->namespace);
			my $pfxprop = $this->_fq2pfx($property->uri);
			
			if ($property->nVals == 1 && $property->firstVal->type == 1)
			{
				$rv .= sprintf("\t\t<%s rdf:resource=\"%s\" />\n", 
				               $pfxprop, 
				               encode_entities_numeric($property->firstVal->toString));
			}
			elsif ($property->nVals == 1 && length $property->firstVal->vocab)
			{
				my $pfxvocab = $this->_fq2pfx($property->firstVal->vocab);
				$rv .= sprintf("\t\t<%s>\n\t\t\t<%s>\n\t\t\t\t<rdf:value>%s</rdf:value>\n\t\t\t</%s>\n\t\t</%s>\n", 
				               $pfxprop, 
				               $pfxvocab,
				               encode_entities_numeric($property->firstVal->toString),
				               $pfxvocab,
				               $pfxprop);
			}
			elsif ($property->nVals == 1)
			{
				$rv .= sprintf("\t\t<%s>%s</%s>\n", 
				               $pfxprop, 
				               encode_entities_numeric($property->firstVal->toString),
				               $pfxprop);
			}
			else
			{
				$rv .= sprintf("\t\t<%s>\n", $pfxprop);
				$rv .= "\t\t\t<rdf:Bag>\n";
				foreach my $val (@{ $property->valObjects })
				{
					if ($val->type == 1)
					{
						$rv .= sprintf("\t\t\t\t<rdf:li rdf:resource=\"%s\" />\n", 
						               encode_entities_numeric($val->toString));
					}
					elsif (length $val->vocab)
					{
						my $pfxvocab = $this->_fq2pfx($val->vocab);
						$rv .= sprintf("\t\t\t\t<rdf:li>\n\t\t\t\t\t<%s>\n\t\t\t\t\t\t<rdf:value>%s</rdf:value>\n\t\t\t\t\t</%s>\n\t\t\t\t</rdf:li>\n", 
						               $pfxvocab, encode_entities_numeric($val->toString), $pfxvocab);
					}
					else
					{
						$rv .= sprintf("\t\t\t\t<rdf:li>%s</rdf:li>\n", 
						               encode_entities_numeric($val->toString));
					}
				}
				$rv .= "\t\t\t</rdf:Bag>\n";
				$rv .= sprintf("\t\t</%s>\n", $pfxprop);
			}
		}
		if (defined $subject->objects)
		{
			foreach my $o (@{ $subject->objects })
			{
				$rv .= sprintf( "\t\t<!-- Also found a perl object: \%s -->\n", ref $o );
			}
		}		
		$rv .= "\t</rdf:Description>\n\n";
	}
	
	$rv .= "</rdf:RDF>\n";
	
	return $rv;

} #/sub dumpRDF



sub parse_rfc2731_namespaces
# Discover namespaces for metadata conforming to RFC 2731. Note
# that eRDF uses the same format for metadata schemas, so we
# can re-use these for eRDF too.
{
	my $this = shift;
	
	# Pull out XML namespaces defined on the root node. In limited circumstances
	# we'll also pull out XML namespaces on other nodes, but not as part of this
	# subroutine.
	foreach my $a ($this->{DOM}->getDocumentElement->getAttributes)
	{
		my $k = $a->getName;
		my $v = $a->getValue;
		next unless $k =~ /^xmlns\:/i;
		$k =~ s/^xmlns\://i;
	
		$this->mergeNS($k, $v, "The '$k' Namespace");
	}
	
	# Find custom RFC 2731 namespaces. FOAF and other stuff goes here.
	foreach my $l ($this->{DOM}->getElementsByTagName('link'))
	{
		my $rel = $l->getAttribute('rel');
		
		if ($rel =~ /^schema[\.\:]([A-Z_][A-Z0-9_-]*)$/i)
		{
			my $prefix = uc($1);
			my $url    = $this->abs_url($l->getAttribute('href'));
			my $title  = $l->getAttribute('title');
			
			$this->mergeNS($prefix, $url, $title || "The '$prefix' Schema");
		}
	}

	# Well-known namespaces
	$this->mergeNS('XHTML', 'http://www.w3.org/1999/xhtml#', 'XHTML' );
	$this->mergeNS('XHV', 'http://www.w3.org/1999/xhtml/vocab#');
	$this->mergeNS('RDF', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'RDF/XML Syntax Specification (Revised)' );
	$this->mergeNS('RDFS', 'http://www.w3.org/2000/01/rdf-schema#', 'RDF Vocabulary Description Language 1.0' );
	$this->mergeNS('DC11', 'http://purl.org/dc/elements/1.1/', 'Dublin Core 1.1' );
	$this->mergeNS('DC', 'http://purl.org/dc/terms/', 'Dublin Core Terms' );
	$this->mergeNS('SCHEMA', 'urn:ietf:rfc:2731#', "The 'Schema' Metaschema" );
	$this->mergeNS('HTTP11', 'urn:ietf:rfc:2616#', 'Hypertext Transfer Protocol 1.1' );
	$this->mergeNS('SERVICE', 'http://www.w3.org/2007/app#', 'Atom Publishing Protocol' );
	$this->mergeNS('OPENID', 'http://openid.net/specs/openid-authentication-1_1.html#', 'OpenID 1.1' );
	
} #/sub parse_rfc2731_namespaces


sub parse_metadata_links
{
	my $this = shift;
	my $myurl = $this->abs_url('');
	
	# Get list of links.
	my @links   = $this->{DOM}->getElementsByTagName('link');
	my @anchors = $this->{DOM}->getElementsByTagName('a');
	push @links, @anchors;
	
	# For each link
	foreach my $l (@links)
	{
		# Subject: This page - SUPERSEDED BY FULL RFDa SUPPORT
		# my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);

		# Subject: Linked page
		my $href = $this->abs_url($l->getAttribute('href'));
		next unless (length $href);
		my $SubjectLinkedPage;
		
		# Handle @title.
		if (length $l->getAttribute('title'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('title'));
			$val->setDom($l);
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/title', $val);
		}

		# Handle @media.
		if (length $l->getAttribute('media'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('media'));
			$val->setDom($l);
			$val->setVocab('http://www.w3.org/1999/xhtml#media');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/format', $val);
		}

		# Handle @title.
		if (length $l->getAttribute('type'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new($l->getAttribute('type'));
			$val->setDom($l);
			$val->setVocab('http://purl.org/dc/terms/IMT');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/format', $val);
		}

		# Handle @lang/@hreflang.
		if (length $l->getAttribute('lang')
		||  length $l->getAttribute('xml:lang')
		||  length $l->getAttribute('hreflang'))
		{
			$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
				unless ($SubjectLinkedPage);

			my $val = Cognition::RDFModel::Value->new(
				$l->getAttribute('hreflang') ||
				$l->getAttribute('xml:lang') ||
				$l->getAttribute('lang')
			);
			$val->setDom($l);
			$val->setVocab('http://purl.org/dc/terms/RFC3066');
			$SubjectLinkedPage->addVal('http://purl.org/dc/terms/language', $val);
		}

		# SUPERSEDED BY FULL RFDa SUPPORT
		## Handle @rel.
		#my $rel   = $l->getAttribute('rel');
		#if (length $rel)
		#{
		#	$rel =~ s/(^\s+|\s+$)//g;
		#	$rel =~ s/\s+/ /g;
		#	my @rel = split / /, $rel;
		#	
		#	my $eRDF = undef;
		#	$eRDF = '-' if (lc($l->tagName) eq 'a');
		#	
		#	foreach my $r (@rel)
		#	{
		#		my $nsrel = $this->_pfx2fq($r, '-', $l);
		#		
		#		my $val = Cognition::RDFModel::Value->new($href, 1);
		#		$val->setDom($l);
		#		$SubjectThisPage->addVal($nsrel, $val);
		#	}
		#}
		#
		## Handle @rev.
		#my $rev   = $l->getAttribute('rev');
		#if (length $rev)
		#{
		#	$rev =~ s/(^\s+|\s+$)//g;
		#	$rev =~ s/\s+/ /g;
		#	my @rev = split / /, $rev;
		#	
		#	my $eRDF = undef;
		#	$eRDF = '-' if (lc($l->tagName) eq 'a');
		#	
		#	$SubjectLinkedPage = Cognition::RDFModel::Subject->new($href)
		#		unless ($SubjectLinkedPage);
		#
		#	foreach my $r (@rev)
		#	{
		#		my $nsrev = $this->_pfx2fq($r, '-', $l);
		#		
		#		my $val = Cognition::RDFModel::Value->new($myurl, 1);
		#		$val->setDom($l);
		#		$SubjectLinkedPage->addVal($nsrev, $val);
		#	}
		#}
		#
		#$this->mergeSubject($SubjectThisPage);
		
		$this->mergeSubject($SubjectLinkedPage);
	}
}


sub parse_metadata_metas
{
	my $this = shift;
	my $myurl = $this->abs_url('');
	
	# Parse <meta>
	foreach my $m ($this->{DOM}->getElementsByTagName('meta'))
	{
		# Value
		my $content = length($m->getAttribute('content')) ?
		                   $m->getAttribute('content') :
		                   Cognition::HTMLParser::Utils::STRINGIFY($m);

		# Subject: This page
		my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);
		
		# @http-equiv
		my $name = $m->getAttribute('http-equiv');
		if (length $name)
		{
			my $nsname = $this->_pfx2fq('HTTP11.'.$name, undef, $m);

			my $val = Cognition::RDFModel::Value->new($content);
			$val->setDom($m);
			$SubjectThisPage->addVal($nsname, $val);
		}
		
		# @name/@property
		else
		{
			$name = $m->getAttribute('property') . ' ' . $m->getAttribute('name');
			next unless (length $name);

			$name =~ s/(^\s+|\s+$)//g;
			$name =~ s/\s+/ /g;
			my @name = split / /, $name;
			
			foreach my $n (@name)
			{
				my $nsname = $this->_pfx2fq($n, undef, $m);
				
				my $val = Cognition::RDFModel::Value->new($content);
				$val->setDom($m);
				if ($m->getAttribute('scheme'))
				{
					my $nsscheme = $this->_pfx2fq($m->getAttribute('scheme'), undef, $m);
					$val->setVocab($nsscheme);
				}
				$SubjectThisPage->addVal($nsname, $val);
			}
		}
		
		$this->mergeSubject($SubjectThisPage);
	}
}


sub parse_metadata_roles
{
	my $this = shift;
	my $myurl = $this->abs_url;
	
	# Parse @role
	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		next unless (length $e->getAttribute('role'));
	
		# Subject: This page
		my $SubjectThisPage = Cognition::RDFModel::Subject->new($myurl);

		# Object: This element
		my $objuri;
		if (length $e->getAttribute('id'))
		{
			$objuri = $myurl . '#' . $e->getAttribute('id');
		}
		else
		{
			$objuri = Cognition::HTMLParser::Utils::XPointer($e, $myurl);
		}

		# Predicate: role
		my $role = $e->getAttribute('role');
		$role =~ s/(^\s+|\s+$)//g;
		$role =~ s/\s+/ /g;
		my @role = split / /, $role;
		
		foreach my $r (@role)
		{
			# Predefined roles from the XHTML vocab namespace
			if ($r =~ /^\:?( banner | contentinfo | definition | main | navigation
				| note | search | secondary | seealso )$/xi)
			{
				$r = "xhv:$1"
			}
			my $nsrole = $this->_pfx2fq($r, undef, $e);
			
			my $val = Cognition::RDFModel::Value->new($objuri, 1);
			$val->setDom($e);
			$SubjectThisPage->addVal($nsrole, $val);
		}		
		
		$this->mergeSubject($SubjectThisPage);
	}
}


sub parse_metadata_erdf
{
	my $this = shift;
	my $myurl = $this->abs_url('');
	
	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		# Value
		my $content = Cognition::HTMLParser::Utils::STRINGIFY($e);

		# Parse eRDF <URL:http://research.talis.com/2005/erdf/wiki/Main/RdfInHtml>.
		if (length ($e->getAttribute('class').$e->getAttribute('rel').$e->getAttribute('rev')))
		{
			# Find eRDF-style parent node. i.e. nearest ancestor with @id set.
			my $parent = $e->getParentNode;
			my $subject = undef;
			my $about = $myurl;
			while (!defined $subject && $parent->nodeType==XML_ELEMENT_NODE)
			{
				if ($parent->getAttribute('id'))
					{ $subject = $parent; }
				else
					{ $parent = $parent->getParentNode(); }
			}
			if ($subject)
			{
				$about .= '#' . $subject->getAttribute('id');
				if (length $subject->getAttribute('href'))
				{
					$about = $this->abs_url($subject->getAttribute('href'));
				}
			}
			my $ParentSubject = Cognition::RDFModel::Subject->new($about);

			my ($MySubject, $addMySubject);
			if (length $e->getAttribute('id'))
			{
				$MySubject = Cognition::RDFModel::Subject->new($myurl.'#'.$e->getAttribute('id'));
			}
			else
			{
				$MySubject = Cognition::RDFModel::Subject->new(Cognition::HTMLParser::Utils::XPointer($e, $this->abs_url));
			}

			# Normalise @class, @ref, @rev and arrayise.
			my @things = ();
			my $class = $e->getAttribute('class');
			$class =~ s/(^\s+|\s+$)//g;
			$class =~ s/\s+/ /g;
			my @class = split / /, $class;
			foreach my $c (@class)
				{ push @things, [('class', $c)]; }
			if ($e->tagName =~ /^(a|area|link)$/i)
			{
				$class = $e->getAttribute('rel');
				$class =~ s/(^\s+|\s+$)//g;
				$class =~ s/\s+/ /g;
				my @class = split / /, $class;
				foreach my $c (@class)
					{ push @things, [('rel', $c)]; }

				$class = $e->getAttribute('rev');
				$class =~ s/(^\s+|\s+$)//g;
				$class =~ s/\s+/ /g;
				my @class = split / /, $class;
				foreach my $c (@class)
					{ push @things, [('rev', $c)]; }
			}

			foreach my $x (@things)
			{
				my ($attr, $c) = @$x;
				
				# rdf:type stuff
				if ($attr eq 'class' && $c =~ /^\-/)
				{
					my $rdftype = $this->_pfx2fq(substr($c,1), '-', $e);
					
					if (($e->tagName eq 'a') && (length $rdftype > length $c))
					{
						my $LinkSubject = Cognition::RDFModel::Subject->new(
							$this->abs_url($e->getAttribute('href'))
						);
						my $val = Cognition::RDFModel::Value->new($rdftype, 1);
						$val->setDom($e);
						$LinkSubject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $val);
						$this->mergeSubject($LinkSubject);
					}
					elsif (length $rdftype > length $c)
					{
						my $val = Cognition::RDFModel::Value->new($rdftype, 1);
						$val->setDom($e);
						$MySubject->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $val);
						$addMySubject++;
					}

					next;
				}
				
				# Only allow namespaced classes.
				my $nsrole = $this->_pfx2fq($c, '-', $e);
				next unless (length $nsrole > length $c);
				
				# Get value.
				my $valString = $e->getAttribute('title') 
					|| Cognition::HTMLParser::Utils::STRINGIFY($e);
				my $valType;
				
				# Resource if it's an <img> element.
				if ($attr eq 'class' && lc($e->tagName) eq 'img')
				{
					$valString = $this->abs_url($e->getAttribute('src'));
					$valType   = 1;
				}
				elsif ($attr eq 'rel' && $e->tagName =~ /^(area|a|link)$/i)
				{
					$valString = $this->abs_url($e->getAttribute('href'));
					$valType   = 1;
				}
				elsif ($attr eq 'rev' && $e->tagName =~ /^(area|a|link)$/i)
				{
					my $LinkSubject = Cognition::RDFModel::Subject->new(
						$this->abs_url($e->getAttribute('href'))
					);
					my $val = Cognition::RDFModel::Value->new($ParentSubject->uri, 1);
					$val->setDom($e);
					$LinkSubject->addVal($nsrole, $val);
					$this->mergeSubject($LinkSubject);
					next;
				}
				elsif ($attr eq 'class' && $e->tagName =~ /^(area|a|link)$/i)
				{
					my $LinkSubject = Cognition::RDFModel::Subject->new(
						$this->abs_url($e->getAttribute('href'))
					);
					my $val = Cognition::RDFModel::Value->new(
						Cognition::HTMLParser::Utils::STRINGIFY($e)
					);
					$val->setDom($e);
					$LinkSubject->addVal($nsrole, $val);
					$this->mergeSubject($LinkSubject);
					next;
				}
				
				# Resource if it has @id.
				if ($attr =~ /^(rel|rev)$/i && length $e->getAttribute('href'))
				{
					my $val = Cognition::RDFModel::Value->new($this->abs_url($e->getAttribute('href')), 1);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
				elsif (length $e->getAttribute('id'))
				{
					my $val = Cognition::RDFModel::Value->new($MySubject->uri, 1);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
				else
				{
					$ParentSubject->setDom($subject) if ($subject);
					my $val = Cognition::RDFModel::Value->new($valString, $valType);
					$val->setDom($e);
					$ParentSubject->addVal($nsrole, $val);
				}
			}
			
			$this->mergeSubject($ParentSubject) if ($ParentSubject->am_i_bovvered);
			$this->mergeSubject($MySubject)     if ($addMySubject);
			
		}
	}
}


sub parse_metadata_rdfa
# http://www.w3.org/TR/rdfa-syntax/#sec_5.5.
{
	my $this = shift;
	my $dom  = shift || $this->{DOM}->documentElement;
	
	return 0 unless ($dom->nodeType == XML_ELEMENT_NODE);
	
	my $fakeid = 0;

	# At the beginning of processing, an initial [evaluation context] is created
	my $base               = shift || $this->abs_url;
	my $parent_subject     = shift || $base;
	my $parent_object      = shift || undef;
	my $uri_mappings       = shift || {};
	my $incomplete_triples = shift || ();
	my $language           = shift || undef;
	
	# Processing begins by applying the processing rules below to the document
	# object, in the context of this initial [evaluation context]. All elements
	# in the tree are also processed according to the rules described below,
	# depth-first, although the [evaluation context] used for each set of rules
	# will be based on previous rules that may have been applied.

	my $current_element = $dom;

	# First, the local values are initialized
	my $recurse            = 1;
	my $skip_element       = 0;
	my $new_subject        = undef;
	my $current_object_resource = undef;
	my $local_uri_mappings = $uri_mappings;
	my $local_incomplete_triples = ();
	my $current_language   = $language;
	
	my $activity = 0;
	
	# Next the [current element] is parsed for [URI mapping]s and these are
	# added to the [local list of URI mappings]. Note that a [URI mapping] 
	# will simply overwrite any current mapping in the list that has the same
	# name
	foreach my $A ($current_element->getAttributes)
	{
		my $attr = $A->getName;
		my $val  = $A->getValue;
		
		if ($attr =~ /^xmlns\:(.*)$/i)
		{
			$local_uri_mappings->{$1} = $this->abs_url($val);
		}
	}
	
	# The [current element] is also parsed for any language information, and
	# if present, [current language] is set accordingly
	if (length $current_element->getAttribute('xml:lang'))
		{ $current_language = $current_element->getAttribute('xml:lang'); }
	elsif (length $current_element->getAttribute('lang'))
		{ $current_language = $current_element->getAttribute('lang'); }
		
	# If the [current element] contains no valid @rel or @rev URI, obtained
	# according to the section on CURIE and URI Processing, then the next step 
	# is to establish a value for [new subject]. Any of the attributes that 
	# can carry a resource can set [new subject]
	my $rel = $current_element->getAttribute('rel');
	$rel =~ s/(^\s+|\s+$)//g;
	$rel =~ s/\s+/ /g;
	my @rel = split / /, $rel;
	my @REL;
	foreach my $r (@rel)
	{
		my $R = $this->_pfx2fq($r, undef, $current_element);
		push @REL, $R if (length $R);
	}
	my $rev = $current_element->getAttribute('rev');
	$rev =~ s/(^\s+|\s+$)//g;
	$rev =~ s/\s+/ /g;
	my @rev = split / /, $rev;
	my @REV;
	foreach my $r (@rev)
	{
		my $R = $this->_pfx2fq($r, undef, $current_element);
		push @REV, $R if (length $R);
	}
	if ((!scalar @REL) && (!scalar @REV))
	{
		if (length $current_element->getAttribute('about'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('about')); }
		elsif (length $current_element->getAttribute('src'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('src')); }
		elsif (length $current_element->getAttribute('resource'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('resource')); }
		elsif (length $current_element->getAttribute('href'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('href')); }
		else
		{
			if ($current_element->tagName eq 'head' || $current_element->tagName eq 'body')
			{
				$new_subject = $this->abs_url;
			}
			elsif (length $current_element->getAttribute('instanceof'))
			{
				$new_subject = Cognition::HTMLParser::Utils::XPointer($current_element, $this->abs_url);
				if (length $current_element->getAttribute('id'))
					{ $new_subject = $this->abs_url.'#'.$current_element->getAttribute('id'); }
			}
			elsif ($parent_object)
			{
				$new_subject = $parent_object;
			}
		}
	}
	
	# If the [current element] does contain a valid @rel or @rev URI, obtained 
	# according to the section on CURIE and URI Processing, then the next step 
	# is to establish both a value for [new subject] and a value for [current
	# object resource]:
	else
	{
		if (length $current_element->getAttribute('about'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('about')); }
		elsif (length $current_element->getAttribute('src'))
			{ $new_subject = $this->abs_url($current_element->getAttribute('src')); }
		else
		{
			if ($current_element->tagName eq 'head' || $current_element->tagName eq 'body')
			{
				$new_subject = $this->abs_url;
			}
			elsif (length $current_element->getAttribute('instanceof'))
			{
				$new_subject = Cognition::HTMLParser::Utils::XPointer($current_element, $this->abs_url);
				if (length $current_element->getAttribute('id'))
					{ $new_subject = $this->abs_url.'#'.$current_element->getAttribute('id'); }
			}
			elsif ($parent_object)
			{
				$new_subject = $parent_object;
			}
		}
		
		if (length $current_element->getAttribute('resource'))
			{ $current_object_resource = $this->abs_url($current_element->getAttribute('resource')); }
		elsif (length $current_element->getAttribute('href'))
			{ $current_object_resource = $this->abs_url($current_element->getAttribute('href')); }
	}
	
	# If in any of the previous steps a [new subject] was set to a non-null
	# value, it is now used to provide a subject for type values
	if ($new_subject && length $current_element->getAttribute('instanceof'))
	{
		my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
		my $VALUE   = Cognition::RDFModel::Value->new(
			$this->_pfx2fq($current_element->getAttribute('instanceof')),
			1
		);
		$SUBJECT->addVal('http://www.w3.org/1999/02/22-rdf-syntax-ns#type', $VALUE);
		$this->mergeSubject($SUBJECT);
		$activity++;
	}
	
	# If in any of the previous steps a [current object resource] was set to
	# a non-null value, it is now used to generate triples
	if ($current_object_resource)
	{
		my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
		foreach my $r (@REL)
		{
			my $VALUE   = Cognition::RDFModel::Value->new($current_object_resource, 1);
			$SUBJECT->addVal($r, $VALUE);
		}
		$this->mergeSubject($SUBJECT);
		$activity++;

		foreach my $r (@REV)
		{
			my $SUBJECT = Cognition::RDFModel::Subject->new($current_object_resource);
			my $VALUE   = Cognition::RDFModel::Value->new($new_subject, 1);
			$SUBJECT->addVal($r, $VALUE);
			$this->mergeSubject($SUBJECT);
			$activity++;
		}
	}
	
	# If however [current object resource] was set to null, but there are 
	# predicates present, then they must be stored as [incomplete triple]s, 
	# pending the discovery of a subject that can be used as the object. Also, 
	# [current object resource] should be set to a newly created [bnode]
	elsif ((scalar @REL) || (scalar @REV))
	{
		foreach my $r (@REL)
		{
			push @$local_incomplete_triples, {
				predicate => $r,
				direction => 'forward'
			};
		}
		foreach my $r (@REV)
		{
			push @$local_incomplete_triples, {
				predicate => $r,
				direction => 'reverse'
			};
		}
		
		$fakeid++;
		$current_object_resource = $this->abs_url . "#fakeid($fakeid)";
	}
	
	# The next step of the iteration is to establish any [current object
	# literal
	my $prop = $current_element->getAttribute('property');
	$prop =~ s/(^\s+|\s+$)//g;
	$prop =~ s/\s+/ /g;
	my @prop = split / /, $prop;
	my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
	my $datatype;
	if (length $current_element->getAttribute('datatype'))
		{ $datatype = $this->_pfx2fq($current_element->getAttribute('datatype'), undef, $current_element); }
	foreach my $p (@prop)
	{	
		my $property = $this->_pfx2fq($p, undef, $current_element);
		next unless (length $property);
		
		my $VALUE;
		if (length $current_element->getAttribute('content'))
		{
			$VALUE = Cognition::RDFModel::Value->new(
				$current_element->getAttribute('content')
			);
		}
		elsif (($current_element->toString !~ /\</) || ($datatype !~ /[A-Z0-9]/i))
		{
			$VALUE = Cognition::RDFModel::Value->new(
				Cognition::HTMLParser::Utils::STRINGIFY($current_element)
			);
		}
		elsif ($datatype eq 'http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral')
		{
			$VALUE = Cognition::RDFModel::Value->new($current_element->toString);
			$VALUE->setDatatype($datatype);
			$recurse = 0;
		}
		else
		{
			$VALUE = Cognition::RDFModel::Value->new(
				Cognition::HTMLParser::Utils::STRINGIFY($current_element)
			);
			$VALUE->setDatatype($datatype);
		}			

		$SUBJECT->addVal($property, $VALUE);
		$this->mergeSubject($SUBJECT);
		$activity++;
	}

	# If the [recurse] flag is 'true', all elements that are children of the 
	# [current element] are processed using the rules described here, using a 
	# new [evaluation context], initialized as follows
	my $flag = 0;
	if ($recurse)
	{
		foreach my $kid ($current_element->childNodes)
		{
			if ($skip_element)
			{
				$flag = $this->parse_metadata_rdfa(
					$kid,
					$base,
					$parent_subject,
					$parent_object,
					$uri_mappings,
					$incomplete_triples,
					$language
				) || $flag;
			}
			else
			{
				$flag = $this->parse_metadata_rdfa(
					$kid,
					$base,
					$new_subject,
					(defined $current_object_resource ? $current_object_resource : ($new_subject ? $new_subject : $parent_subject)),
					$local_uri_mappings,
					$local_incomplete_triples,
					$current_language
				) || $flag;
			}
		}	
	}
	
	# If the [skip element] flag is 'false', and either: the previous step
	# resulted in a 'true' flag, or [new subject] was set to a non-null and
	# non-bnode value, then any [incomplete triple]s within the current context
	# should be completed:
	if (!$skip_element && ($flag || ((defined $new_subject) && ($new_subject !~ /\#fakeid/))))
	{
		# Loop through list of incomplete triples...
		foreach my $it (@$incomplete_triples)
		{
			my $direction = $it->{direction};
			my $predicate = $it->{predicate};
			
			if ($direction eq 'forward')
			{
				my $SUBJECT = Cognition::RDFModel::Subject->new($parent_subject);
				my $VALUE   = Cognition::RDFModel::Value->new($new_subject, 1);
				$SUBJECT->addVal($predicate, $VALUE);
				$this->mergeSubject($SUBJECT);
				$activity++;
			}
			else
			{
				my $SUBJECT = Cognition::RDFModel::Subject->new($new_subject);
				my $VALUE   = Cognition::RDFModel::Value->new($parent_subject, 1);
				$SUBJECT->addVal($predicate, $VALUE);
				$this->mergeSubject($SUBJECT);
				$activity++;
			}
		}
	}
	
	return 1 if ($activity || $new_subject || $flag);
	return 0;
}


sub parse_uformats
{
	my $this = shift;
	
	$this->{uF}->{hCard} = [ Cognition::uF::hCard::parse_all($this) ];
	foreach my $hcard (@{ $this->{uF}->{hCard} })
		{ $this->mergeSubject( $hcard->rdf_subject ); }
	if (defined $this->{uF}->{hCard}->[0])
		{ $this->mergeNS( $this->{uF}->{hCard}->[0] ); }

	$this->{uF}->{geo} = [ Cognition::uF::geo::parse_all($this) ];
	foreach my $geo (@{ $this->{uF}->{geo} })
		{ $this->mergeSubject( $geo->rdf_subject ); }
	if (defined $this->{uF}->{geo}->[0])
		{ $this->mergeNS( $this->{uF}->{geo}->[0] ); }

	$this->{uF}->{adr} = [ Cognition::uF::adr::parse_all($this) ];
	foreach my $adr (@{ $this->{uF}->{adr} })
		{ $this->mergeSubject( $adr->rdf_subject ); }
	if (defined $this->{uF}->{adr}->[0])
		{ $this->mergeNS( $this->{uF}->{adr}->[0] ); }

	$this->{uF}->{Tag} = [ Cognition::uF::RelTag::parse_all($this) ];
	foreach my $tag (@{ $this->{uF}->{Tag} })
	{
		$this->mergeSubject( $tag->rdf_subject );
		my $val = Cognition::RDFModel::Value->new($tag->uri, 1);
		my $url = $this->abs_url;
		$this->{RDF}->{$url}->addVal('http://microformats.org/wiki/rel-tag#tag', $val);
	}
	if (defined $this->{uF}->{Tag}->[0])
		{ $this->mergeNS( $this->{uF}->{Tag}->[0] ); }
	
	$this->{uF}->{License} = [ Cognition::uF::RelLicense::parse_all($this) ];
	foreach my $license (@{ $this->{uF}->{License} })
	{
		$this->mergeSubject( $license->rdf_subject );
		my $val = Cognition::RDFModel::Value->new($license->uri, 1);
		my $url = $this->abs_url;
		$this->{RDF}->{$url}->addVal('http://web.resource.org/cc/license', $val);
	}
	if (defined $this->{uF}->{License}->[0])
		{ $this->mergeNS( $this->{uF}->{License}->[0] ); }
	
	$this->{uF}->{Enclosure} = [ Cognition::uF::RelEnclosure::parse_all($this) ];
	foreach my $Enclosure (@{ $this->{uF}->{Enclosure} })
	{
		$this->mergeSubject( $Enclosure->rdf_subject );
		my $val = Cognition::RDFModel::Value->new($Enclosure->uri, 1);
		my $url = $this->abs_url;
		$this->{RDF}->{$url}->addVal('http://microformats.org/wiki/rel-enclosure#enclosure', $val);
	}
	if (defined $this->{uF}->{Enclosure}->[0])
		{ $this->mergeNS( $this->{uF}->{Enclosure}->[0] ); }
	
	$this->{uF}->{hCalendar} = [ Cognition::uF::hCalendar::parse_all($this) ];
	foreach my $hcalendar (@{ $this->{uF}->{hCalendar} })
	{
		$this->mergeSubject( $hcalendar->rdf_subject );

		foreach my $childtype (qw(vevent vtodo valarm vfreebusy))
		{
			foreach my $h ( @{ $hcalendar->{$childtype} } )
			{
				$this->mergeSubject( $h->rdf_subject );
			}
		}
	}
	if (defined $this->{uF}->{hCalendar}->[0])
		{ $this->mergeNS( $this->{uF}->{hCalendar}->[0] ); }

	# hAtom
	# hReview ?

} #/sub parse_uformats


sub parse_doc_structure
{

	my $this = shift;
	my $rv   = { kids=>() };
	my $lvl  = 1;
	my $ptr  = $rv;
	my @nodelist;
	
	my $count_lists;
	my $count_tables;
	my $count_figures;

	foreach my $e ($this->{DOM}->getElementsByTagName('*'))
	{
		my $tag = $e->localname;
		
		if ($tag eq 'figure' || $e->getAttribute('class') =~ / (^|\s) figure (\s|$) /x)
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::Figure';
			push @{ $ptr->{kids}->[-1]->{kids} }, $node;
			$node->{tag} = 'figure';
			$node->{count} = ++$count_figures;
			$node->{object} = Cognition::uF::figure::parse($this, $e);
			$node->{_dom} = $e;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			push @nodelist, $node;
		}
		
		elsif ($e->getAttribute('class') =~ / (^|\s) xoxo (\s|$) /x)
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::XOXO';
			push @{ $ptr->{kids}->[-1]->{kids} }, $node;
			$node->{tag} = 'xoxo';
			$node->{count} = ++$count_lists;
			$node->{object} = Cognition::uF::xoxo::parse($this, $e);
			$node->{_dom} = $e;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			$node->{heading} = length $e->getAttribute('title') ? $e->getAttribute('title') : undef;
			push @nodelist, $node;
		}

		elsif ($tag =~ /^h([1-6])$/i)
		{
			my $thisLevel = $1;
			my $node = {};
			bless $node, 'Cognition::DocModel::Section';
			
			while ($thisLevel > $lvl)
			{
				my $parent = $ptr->{kids}->[-1];
				$parent->{up} = $ptr;
				$ptr = $parent;
				$lvl++;
			}
			while ($thisLevel < $lvl)
			{
				$lvl--;
				$ptr = $ptr->{up};
			}
			if ($thisLevel == $lvl)
			{
				push @{ $ptr->{kids} }, $node;
				$node->{up} = $ptr;
			}
			
			$node->{tag}     = $tag;
			$node->{_dom}    = $e;
			$node->{_page}   = $this;
			$node->{outline} = Cognition::HTMLParser::Utils::outline_point($tag);
			$node->{heading} = Cognition::HTMLParser::Utils::STRINGIFY($e);
			push @nodelist, $node;
		}
		
		elsif ( $tag eq 'table' && (length $e->getAttribute('summary') || defined $e->getElementsByTagName('caption')->item(0)) )
		{
			my $node = {};
			bless $node, 'Cognition::DocModel::Table';
			push @{ $ptr->{kids}->[-1]->{kids} }, $node;
			$node->{tag} = $tag;
			$node->{_dom} = $e;
			$node->{summary} = $e->getAttribute('summary');
			$node->{count} = ++$count_tables;
			$node->{_page} = $this;
			$node->{up} = $ptr->{kids}->[-1];
			my @captions = $e->getElementsByTagName('caption');
			if (@captions)
			{
				$node->{caption} = Cognition::HTMLParser::Utils::STRINGIFY($captions[0]);
			}
			push @nodelist, $node;
		}
	}

	foreach my $N (@nodelist)
	{
		$this->mergeSubject($N->rdf_subject);
		$this->mergeNS($N);
	}

	$this->{Structure}   = $rv;
	return $rv;	

} #/sub parse_doc_structure


sub parse_external_rdf
{
	my $this = shift;
	
	return unless ($this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta'));
		
	foreach my $val ( @{ $this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/1999/xhtml/vocab#meta')->valObjects } )
	{
		my $request  = HTTP::Request->new(GET => $val->toString);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(rdf)/i)
		{
			$this->parse_rdf_data(
				$response->content,
				$response->base,
				$response->header('content-type')
				);
		}
	}
} #/sub parse_external_rdf


sub parse_grddl
{
	my $this = shift;

	my @transformations;
	my $prop = $this->{RDF}->{ $this->abs_url }
		->property('http://www.w3.org/2003/g/data-view#transformation');
	if ($prop)
	{
		foreach my $v (@{$prop->valObjects})
		{
			push @transformations, $v->toString if ($v->type);
		}
	}
	
	my $attr = $this->_fq2pfx('http://www.w3.org/2003/g/data-view#transformation');
	my $attv = $this->{DOM}->documentElement->getAttribute($attr);
	$attv =~ s/(^\s+|\s+$)//g;
	$attv =~ s/\s+/ /g;
	my @attv = split / /, $attv;
	foreach my $a (@attv)
		{ push @transformations, $this->abs_url($a); }
	
	foreach my $t (@transformations)
	{
		my $request  = HTTP::Request->new(GET => $t);
		my $response = $this->{UA}->request($request);

		if ($response->header('content-type') =~ /(xslt|xml)/i)
		{
			my $parser     = XML::LibXML->new();
			my $xslt       = XML::LibXSLT->new();
			my $style_doc  = $parser->parse_string($response->content);
			my $stylesheet = $xslt->parse_stylesheet($style_doc);
			my $results    = $stylesheet->transform($this->{DOM});
			my $rdfxml     = $stylesheet->output_string($results);
			$this->parse_rdf_data($rdfxml);
		}
	}
}


sub parse_rdf_data
{
	my $this = shift;
	my $rdf  = shift;
	my $base = shift || $this->{RESPONSE}->base;
	my $mime = shift || 'application/rdf+xml';
	
	my $baseobj = new RDF::Redland::URI("$base");
	my $parser  = new RDF::Redland::Parser(undef, $mime);
	my $stream  = $parser->parse_string_as_stream($rdf, $baseobj);

	while ($stream && !$stream->end)	
	{
		my $statement = $stream->current;
		
		my $subj = $statement->subject;
		my $prop = $statement->predicate;
		my $obj  = $statement->object;
		
		if ($subj->is_resource && $prop->is_resource)
		{
			my $S = Cognition::RDFModel::Subject->new($subj->uri->as_string);
			my $P = $prop->uri->as_string;
			my $V;
			
			if ($obj->is_resource)
				{ $V = Cognition::RDFModel::Value->new($obj->uri->as_string, 1); }
			else
				{ $V = Cognition::RDFModel::Value->new($obj->literal_value); }

			$S->addVal($P, $V);				
			$this->mergeSubject($S);	
		}
		
		$stream->next;
	}
	
	if ($RDF::Redland::VERSION gt '1.00_06_00')
	{
		my %ns = $parser->namespaces_seen;
		foreach my $k (keys %ns)
		{
			$this->mergeNS($k, $ns{$k}->as_string, "RDF $k");
		}
	}
	else
	{
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\"([^"]*)\"/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix");
		}
		while ($rdf =~ /xmlns\:?([A-Za-z0-9_-]*)\=\'([^']*)\'/g)
		{
			my $nsurl  = $2;
			my $prefix = length($1) ? $1 : 'defaultpfx';
			$this->mergeNS($prefix, $nsurl, "RDF $prefix");
		}
	}
} #/sub parse_rdf_data


sub parse_metadata
{
	my $this = shift;
	my $myurl = $this->abs_url('');

	$this->parse_rfc2731_namespaces;

	# Parse <title>.
	foreach my $t ($this->{DOM}->getElementsByTagName('title'))
	{
		my $S = Cognition::RDFModel::Subject->new($myurl);
		my $val = Cognition::RDFModel::Value->new(Cognition::HTMLParser::Utils::STRINGIFY($t));
		$val->setDom($t);
		$S->addVal('http://purl.org/dc/terms/title', $val);
		$this->mergeSubject($S);
	}

	$this->parse_metadata_links;
	$this->parse_metadata_metas;
	$this->parse_metadata_roles;
	
	$this->parse_metadata_erdf
		if ($this->{options}->{p_erdf});
	
	$this->parse_metadata_rdfa
		if ($this->{options}->{p_rdfa});

	$this->parse_uformats
		if ($this->{options}->{p_uf});
	
	$this->parse_doc_structure
		if ($this->{options}->{p_structure});
	
	$this->parse_external_rdf
		if ($this->{options}->{p_rdf});

	$this->parse_grddl
		if ($this->{options}->{p_grddl});
		
} #/sub parse_metadata


sub _pfx2fq
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift;
	my $dom  = shift;
	my ($ns, $term, $nsurl);
	
	$mode = '.:' unless length $mode;
	
	if (defined $dom && ($mode eq '.:') && length ($dom->getAttribute('rel').$dom->getAttribute('rev')))
	{
		# From HTML 4.01 and HTML 3.2 specs
		if ($key =~ /^\:?( Alternate | Stylesheet | Start | Next | Prev |
			Contents | Index | Glossary | Chapter | Section | Subsection |
			Appendix | Help | Bookmark | Previous | Search | Made )$/ix)
		{
			$key = "xhv:$key";
		}
		# And the HTML 5 spec.
		elsif ($key =~ /^\:?( alternate | archives | author | bookmark | contact |
			external | feed | first | help | icon | index | last | license | next |
			nofollow | noreferrer | pingback | prefetch | prev | search |
			stylesheet | sidebar | tag | up )$/ix)
		{
			$key = "xhv:$key";
		}
		# Some others. Not sure where they're defined but RDFa test suite mentions.
		elsif ($key =~ /^\:?( cite | meta | p3pv1 )$/ix)
		{
			$key = "xhv:$key";
		}
		elsif ($key eq 'transformation')
		{
			if (!defined $this->{NS}->{grddl})
			{
				my $pfx = $this->mergeNS('grddl', 'http://www.w3.org/2003/g/data-view#', 'GRDDL');
				$key = "$pfx:$key";
			}
		}
	}
	
	# In "dash mode" allows the hyphen to be used as a namespace
	# prefix separator. This is for eRDF.
	if (($mode eq '-' && $key =~ /^([A-Z_][A-Z0-9_]*)[\.\:-](.+)$/i)
	# Normally, dash is allowed as part of the namespace prefix.    
	|| ($mode eq '.:' && $key =~ /^([A-Z_][A-Z0-9_-]*)[\.\:](.+)$/i))
	{
		$ns = $1;
		$term = $2;
		$nsurl = $this->{NS}->{uc($ns)}->{nsurl};
		
		# If we haven't found the namespace yet, then check xmlns:* attributes.
		if ($dom && !length $nsurl)
		{
			$nsurl = Cognition::HTMLParser::Utils::xmlns($ns, $dom);
			$this->mergeNS($ns, $nsurl, $ns);
		}
		
		# In eRDF, cope with undefined prefixes by assuming that
		# the class name was never intended as eRDF.
		if ($mode eq '-' && !$nsurl)
		{
			return $key;
		}
		
		# Otherwise, if an undefined prefix is used, complain quietly.
		$nsurl = 'http://undefined-namespace-prefix.invalid/' if (!$nsurl);
		
		return $nsurl.$term;
	}

	# If no valid prefix is found, treat as case-insensitive default namespace
	return lc($key);

} #/sub _pfx2fq


sub _fq2pfx
{
	my $this = shift;
	my $key  = shift;
	my $mode = shift || ':';

	$key =~ s/(^\<|\>$)//g;
	$key =~ /^(.*[\#\/])([^\#\/]+)$/;

	my $nsurl = $1;
	$key      = $2;
	my $pfx;

	$key = 'null' unless (length $key);

	foreach my $k (keys %{ $this->{NS} })
	{
		if ($this->{NS}->{$k}->{nsurl} eq $nsurl)
		{
			$pfx = lc($k);
			return $pfx.$mode.$key;
		}
	}

	return $key;

} #/sub _fq2pfx


sub abs_url
{
	my $this  = shift;
	my $param = shift || '';
	my $dom   = shift;
	
	# Support for CURIEs.
	if ($param =~ /^\[([A-Za-z0-9_]+)\:([A-Za-z0-9_]+)\]$/)
	{	
		my $ns    = $1;
		my $curie = $2;

		my $expanded = $this->_pfx2fq("$ns:$curie", undef, $dom);
		return $expanded if (length $expanded);
	}
	
	my $url = url $param, $this->{BASE};
	return $url->abs->as_string;
} #/sub abs_url


sub mergeSubject
{
	my $this    = shift;
	my $mergeme = shift;
	my $target;

	return unless (defined $mergeme);
	
	foreach my $ex (keys %{$this->{RDF}})
	{
		my $existing = $this->{RDF}->{$ex};
		next unless (defined $existing);
		
		if ($existing->equal_to($mergeme))
		{
			$target = $ex;
			last;
		}
	}
	
	if (defined $target)
	{
		$this->{RDF}->{$target}->merge($mergeme);
	}
	else
	{
		$this->{RDF}->{ $mergeme->uri } = $mergeme;
	}
}


sub mergeNS
# Can be passed either an object with a metadata_ns method or
# a list of suggestedPrefix, url, title.
{
	my $this    = shift;
	my $mergeme = shift;
	
	if (ref $mergeme && UNIVERSAL::can($mergeme, 'metadata_ns'))
	{
		my $mns = $mergeme->metadata_ns;
		foreach my $k (keys %$mns)
		{
			$this->mergeNS($k, $mns->{$k}->{nsurl}, $mns->{$k}->{title});
		}
	}
	
	else
	{
		my $suggestedPrefix = uc($mergeme);
		my $url             = shift;
		my $title           = shift;
		
		if (defined $this->{NS}->{$suggestedPrefix}
		&& $this->{NS}->{$suggestedPrefix}->{nsurl} eq $url)
			{ return $suggestedPrefix; }
			
		else
		{
			foreach my $k (keys %{ $this->{NS} })
			{
				return $k if ($this->{NS}->{$k}->{nsurl} eq $url);
			}
			
			while (defined $this->{NS}->{$suggestedPrefix}->{nsurl})
			{
				$suggestedPrefix = sprintf('ns%03d', int(rand(1000)));
			}
			
			$this->{NS}->{$suggestedPrefix} = {
				nsurl => $url,
				title => $title
			};
			return $suggestedPrefix;
		}
	}
}


sub html2xhtml
{
	my $this = shift;
	my $html = shift;
	
	$html =~ s/\<\!DOCTYPE[^>]*\>//i;
	
	my $t = HTML::TreeBuilder->new();
	$t->implicit_tags(1);
	$t->p_strict(1);
	$t->parse_content($html);
	$t->{'xmlns'} = 'http://www.w3.org/1999/xhtml';
	
	my $xhtml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
		. "<!DOCTYPE html\n"
		. "\tPUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		. "\t\"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
		. $t->as_XML(undef, "\t", {
			'area'=>1,
			'base'=>1,
			'basefont'=>1,
			'br'=>1,
			'col'=>1,
			'frame'=>1,
			'hr'=>1,
			'img'=>1,
			'input'=>1,
			'isindex'=>1,
			'link'=>1,
			'meta'=>1,
			'param'=>1
		});
	
	# XHTML now has some annoying named entities.
	#my %ents = %HTML::Entities::entity2char;
	#foreach my $e (keys %ents)
	#{
	#	my $char = $ents{$e};
	#	$e =~ s/\;$//;
	#	$xhtml =~ s/ \& $e \; / sprintf "&#x%X;", ord($char) /gex;
	#}
	
	return $xhtml;
} #/sub html2xhtml


1; #/package Cognition::HTMLParser




